# (c) Copyright 2018 Trent Hauck
# All Rights Reserved

from pathlib import Path
from typing import Dict

import torch
from Bio.SeqRecord import SeqRecord

from gcgc.parser import SequenceParser


class TorchSequenceParser(SequenceParser):
    def parse_record(self, input_seq: SeqRecord, path: Path) -> Dict:
        """
        Convert the incoming SeqRecord to a dictionary of features.
        """

        parsed_features = super().parse_record(input_seq, path)
        parsed_features["seq_tensor"] = torch.LongTensor(parsed_features["seq_tensor"])

        parsed_features["id"] = input_seq.id

        if self.has_file_features:
            for ff in self.file_features:
                parsed_features[ff.name] = torch.tensor(parsed_features[ff.name])

        return parsed_features



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved

import tempfile
import unittest
from pathlib import Path

from gcgc.datasets import dataset


class TestDataset(unittest.TestCase):
    def test_uniprot_dataset(self):

        with tempfile.TemporaryDirectory() as tmpdirname:
            tmpdir_path = Path(tmpdirname)
            taxon_dataset = dataset.TaxonDataset([83333], tmpdir_path)
            taxon_dataset.download_files()

            files = list(tmpdir_path.glob("*.fasta"))

            self.assertEqual(len(files), 1)
            self.assertTrue(all(f.stat().st_size > 0 for f in files))



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved

from Bio.Alphabet import IUPAC

from gcgc.alphabet.base import EncodingAlphabet


class ExtendedIUPACProteinEncoding(EncodingAlphabet, IUPAC.ExtendedIUPACProtein):
    """
    Implements an encoding alphabet using the IUPAC extended protein letters.
    """


class ExtendedIUPACDNAEncoding(EncodingAlphabet, IUPAC.ExtendedIUPACDNA):
    """
    Implements an encoding alphabet using the IUPAC extended dna letters.
    """


class IUPACUnambiguousDNAEncoding(EncodingAlphabet, IUPAC.IUPACUnambiguousDNA):
    """
    Implements an encoding alphabet using the IUPAC unambiguous dna letters.
    """


class IUPACUnambiguousRNAEncoding(EncodingAlphabet, IUPAC.IUPACUnambiguousRNA):
    """
    Implements an encoding alphabet using the IUPAC unambiguous rna letters.
    """


class IUPACAmbiguousDNAEncoding(EncodingAlphabet, IUPAC.IUPACAmbiguousDNA):
    """
    Implements an encoding alphabet using the IUPAC unambiguous dna letters.
    """


class IUPACAmbiguousRNAEncoding(EncodingAlphabet, IUPAC.IUPACAmbiguousRNA):
    """
    Implements an encoding alphabet using the IUPAC unambiguous rna letters.
    """


class IUPACProteinEncoding(EncodingAlphabet, IUPAC.IUPACProtein):
    """
    Implements an encoding alphabet using the IUPAC protein letters.
    """



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved

from dataclasses import dataclass


@dataclass
class Field:
    """
    The base class for a Field, this holds the name.
    """

    name: str



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved

from dataclasses import dataclass
from typing import Dict, List, Optional

from gcgc.encoded_seq import EncodedSeq
from gcgc.parser.gcgc_record import GCGCRecord
from gcgc.exceptions import EncodedSeqLengthParserException
from gcgc.fields import FileMetaDataField
from gcgc.fields import AnnotationField


@dataclass
class EncodedSeqLengthParser:
    pad_to: Optional[int] = None
    conform_to: Optional[int] = None

    def parse_encoded_seq_record(self, encoded_seq: EncodedSeq) -> EncodedSeq:
        num_non_none = sum(1 for x in [self.pad_to, self.conform_to] if x is not None)
        if num_non_none > 1:
            raise EncodedSeqLengthParserException("One one of pad_to or conform_to an be set.")

        if self.pad_to:
            return encoded_seq.pad(pad_to=self.pad_to)

        if self.conform_to:
            return encoded_seq.conform(conform_to=self.conform_to)

        raise RuntimeError(
            "Could not parse seq length for some reason, check parser configuration."
            f"conform_to={self.conform_to} pad_to={self.pad_to}"
        )


@dataclass
class SequenceParser:
    """
    A configurable sequence parser.
    """

    encapsulate: bool = True
    seq_length_parser: Optional[EncodedSeqLengthParser] = None
    file_features: Optional[List[FileMetaDataField]] = None
    annotation_features: Optional[List[AnnotationField]] = None

    def parse_record(self, gcgc_record: GCGCRecord) -> Dict:
        """
        Convert the incoming SeqRecord to a dictionary of features.
        """

        parsed_features: Dict = {}

        if self.encapsulate:
            es = EncodedSeq.from_seq(gcgc_record.seq_record.seq).encapsulate()

        if self.seq_length_parser:
            es = self.seq_length_parser.parse_encoded_seq_record(es)

        parsed_features["id"] = gcgc_record.seq_record.id
        parsed_features["seq_tensor"] = es.integer_encoded
        parsed_features.update(self._generate_file_features(gcgc_record.path))

        return parsed_features

    @property
    def has_file_features(self) -> bool:
        return bool(self.file_features)

    def _generate_file_features(self, path):
        file_features = {}

        if self.has_file_features:
            for ff in self.file_features:
                file_features[ff.name] = ff.encode(path)

        return file_features

    @property
    def has_annotation_features(self) -> bool:
        return bool(self.annotation_features)

    def _generate_annotation_features(self, seq_record):
        seq_records_features = {}

        if self.has_seq_records_features:
            for ff in self.seq_records_features:
                seq_records_features[ff.name] = ff.encode(seq_record.annotations)

        return seq_records_features



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved

from dataclasses import dataclass

from pathlib import Path
from Bio.SeqRecord import SeqRecord


@dataclass
class GCGCRecord:
    path: Path
    seq_record: SeqRecord



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved

from typing import Sequence

import numpy as np
from Bio.Seq import Seq

from gcgc.alphabet.base import EncodingAlphabet
from gcgc.alphabet.utils import biopython_alphabet_to_gcgc_alphabet
from gcgc.exceptions import GCGCAlphabetException


class EncodedSeq(Seq):
    def __init__(self, data, alphabet) -> None:
        if not isinstance(alphabet, EncodingAlphabet):
            raise GCGCAlphabetException(f"Cannot use alphabet of type {type(alphabet)}.")

        super().__init__(data, alphabet)

    @classmethod
    def from_seq(cls, seq: Seq):
        """
        Instantiate an EncodedSeq object from a Seq object.
        """

        gcgc_alphabet = biopython_alphabet_to_gcgc_alphabet(seq.alphabet)
        return cls(str(seq), gcgc_alphabet)

    def encapsulate(self) -> "EncodedSeq":
        return self.alphabet.START + self + self.alphabet.END

    def pad(self, pad_to: int = 50) -> "EncodedSeq":
        """
        Pad a sequence up to `pad_to` characters.
        """

        seq_len = len(self)

        if seq_len < pad_to:
            n_extra_chars = pad_to - seq_len
            extra_chars = self.alphabet.PADDING * n_extra_chars
        else:
            extra_chars = ""

        return self + extra_chars

    def conform(self, conform_to: int = 50) -> "EncodedSeq":
        seq_len = len(self)

        if seq_len == conform_to:
            return self
        elif seq_len < conform_to:
            return self.pad(pad_to=conform_to)
        else:
            return self[:conform_to]

    @property
    def integer_encoded(self):
        return self.alphabet.integer_encode(self)

    @property
    def one_hot_encoded(self) -> Sequence[Sequence[int]]:
        """
        Encodes D x N where D is the size of the alphabet and N is the padding.

        Returns:
            A one hot encoded matrix representing the sequence.
        """

        encoded_sequence = self.alphabet.integer_encode(self)
        encoded_len = len(encoded_sequence)
        letters_len = len(self.alphabet.letters_and_tokens)

        one_hot_seq = np.zeros((encoded_len, letters_len), dtype=np.int)
        one_hot_seq[np.arange(encoded_len), encoded_sequence] = 1

        return one_hot_seq.tolist()

    def __add__(self, other) -> "EncodedSeq":
        """
        Add two enccoded sequences together.
        """
        added_seq = super().__add__(other)
        return self.from_seq(added_seq)

    def __getitem__(self, index) -> "EncodedSeq":
        got_item = super().__getitem__(index)
        if isinstance(index, int):
            return got_item
        else:
            return self.from_seq(got_item)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved

import unittest

from Bio.Alphabet import IUPAC

from gcgc.alphabet import IUPACUnambiguousDNAEncoding
from gcgc.alphabet.utils import biopython_alphabet_to_gcgc_alphabet


class TestUtils(unittest.TestCase):
    def test_biopython_alphabet_to_gcgc_alphabet(self):
        dna = IUPAC.IUPACUnambiguousDNA()

        klass = biopython_alphabet_to_gcgc_alphabet(dna)
        self.assertIsInstance(klass, IUPACUnambiguousDNAEncoding)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved


class GCGCAlphabetException(Exception):
    """
    Raised when there's an error with GCGC's alphabet.
    """


class GCGCAlphabetLetterEncodingException(GCGCAlphabetException):
    """
    Thrown when GCGC's alphabet tries to encode a letter it wasn't aware of.
    """


class GCGCAlphabetLetterDecodingException(GCGCAlphabetException):
    """
    Thrown when GCGC's alphabet tries to decode a letter it wasn't aware of.
    """


class EncodedSeqLengthParserException(Exception):
    """
    Raised when there's an issue with the EncodedSeqLengthParser's parsing.
    """



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved

import unittest
from pathlib import Path

from Bio.Alphabet import IUPAC
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from gcgc.parser.gcgc_record import GCGCRecord
from gcgc.fields import FileMetaDataField
from gcgc.parser.base import EncodedSeqLengthParser, SequenceParser


class TestParser(unittest.TestCase):
    def test_parser(self):
        vocab = [Path("ecoli"), Path("human")]

        f = FileMetaDataField.from_paths("species", vocab)
        ff = [f]

        length_parser = EncodedSeqLengthParser(conform_to=10)

        sp = SequenceParser(encapsulate=True, seq_length_parser=length_parser, file_features=ff)

        dna = IUPAC.IUPACUnambiguousDNA()
        input_seq = SeqRecord(Seq("ATCG", alphabet=dna))

        test_values = [
            (input_seq, Path("ecoli"), 0),
            (input_seq, Path("human"), 1),
            (input_seq, Path("human"), 1),
        ]

        for i, p, es in test_values:
            r = GCGCRecord(path=p, seq_record=i)
            resp = sp.parse_record(r)
            self.assertEqual(resp["species"], es)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved

import asyncio
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Sequence

import aiohttp

logger = logging.getLogger(__name__)


@dataclass
class DataFile:
    """
    A an external file download and place at the local path.
    """

    url: Optional[str]
    local_path: Path

    async def download(self, session):
        """
        Given a session, query the url.

        Args:
            session: The asyncio session.

        Returns:
            The released response.
        """

        async with session.get(self.url) as response:
            with open(self.local_path, "wb") as f_handle:
                while True:
                    chunk = await response.content.read(1024)
                    if not chunk:
                        break
                    f_handle.write(chunk)

            logger.info(f"Downloaded {self.url} to {self.local_path}.")
            return await response.release()


DataFiles = List[DataFile]


class UniprotDataset(object):
    """
    A base class that loads data from Uniprot.
    """

    def __init__(self, data_files: DataFiles, file_format: str = "fasta"):
        """
        Initialize the UniprotDataset object.

        Args:
            data_files: The list for FileToDownload object.
            file_format: The file format to use.
        """

        self.data_files = data_files
        self.file_format = file_format

    @staticmethod
    def _url(query_str, file_format):
        return f"https://www.uniprot.org/uniprot/?{query_str}&format={file_format}"

    def download_files(self):
        asyncio.run(self._download())

    async def _download(self):
        async with aiohttp.ClientSession() as session:
            downloaded_files = [f.download(session) for f in self.data_files]
            await asyncio.gather(*downloaded_files)


class TaxonDataset(UniprotDataset):
    """
    Download a set of taxons IDs from uniprot.
    """

    def __init__(
        self, taxon_ids: Sequence[int], data_directory: Path = Path("."), file_format: str = "fasta"
    ):
        """
        Initialize the TaxonDataset object.

        Args:
            taxon_ids: A list of integers representing taxon ids.
            data_directory: The data directory path for the taxon.
            file_format: The file format to use for the downloaded data.
        """

        self.taxon_ids = taxon_ids
        files_to_download: DataFiles = []

        for taxon_id in self.taxon_ids:

            url = self._url(f"query=reviewed:yes+AND+organism:{taxon_id}", file_format=file_format)
            local_path = data_directory / str(f"{taxon_id}.{file_format}")

            files_to_download.append(DataFile(url, local_path))

        super().__init__(files_to_download, file_format)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved

import unittest

from gcgc.alphabet.iupac import IUPACProteinEncoding
from gcgc.parser import SequenceParser
from gcgc.tests.fixtures import ECOLI_PATH, P53_HUMAN
from gcgc.third_party.pytorch_utils.data import GenomicDataset


class TestPyTorchUtils(unittest.TestCase):
    def setUp(self) -> None:
        self.sp = SequenceParser()

    def test_load_dataset(self):
        def yielder():
            yield P53_HUMAN

        test_dataset = GenomicDataset.from_paths(yielder(), self.sp, "fasta")
        self.assertEqual(len(test_dataset), 1)

    def test_index_multiple_files(self):

        glob = ECOLI_PATH.glob("*.fasta")

        pe = IUPACProteinEncoding()
        test_dataset = GenomicDataset.from_paths(glob, self.sp, "fasta", pe)
        self.assertEqual(len(test_dataset), 25)

        test_sequences = {
            0: "sp|C5A0C3|ARGE_ECOBW",
            1: "sp|P23908|ARGE_ECOLI",
            24: "sp|B1X825|LFTR_ECODH",
        }

        for idx, expected_id in test_sequences.items():
            actual_record = test_dataset[idx]
            self.assertEqual(actual_record["id"], expected_id)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved

from abc import ABC
from typing import Sequence

from gcgc.exceptions import GCGCAlphabetLetterEncodingException


class EncodingAlphabet(ABC):
    """
    The Encoding Alphabet is meant to be a baseclass for other alphabets. This is similar to
    vocabularies often found in NLP packages, though the extent of the vocabulary is known upfront.
    """

    START = ">"
    END = "<"
    PADDING = "|"

    def __init__(self):
        """
        Create the EncodingAlphabet object.
        """

        self.letters_and_tokens = self.letters + self.START + self.END + self.PADDING

        self.encoding_index = {letter: idx for idx, letter in enumerate(self.letters_and_tokens)}
        self.decoding_index = {idx: letter for letter, idx in self.encoding_index.items()}

    def __len__(self):
        return len(self.letters_and_tokens)

    def encode_token(self, token: str) -> int:
        """
        Given a particular token, return the integer representation.

        Args:
            token: The token to convert.

        Returns:
            The integer representing the token.
        """

        return self.encoding_index[token]

    def decode_token(self, int_token: int) -> str:
        """
        Decode a token. This is the inverse of encode_token.

        Args:
            int_token: The integer representation of a token.

        Returns:
            The str which was encoded by the integer.
        """

        return self.decoding_index[int_token]

    def integer_encode(self, seq: str) -> Sequence[int]:
        """
        Integer encode the sequence.

        Args:
            seq: The sequence to pad.

        Returns:
            The integer sequence representation of the sequence.
        """

        try:
            return [self.encoding_index[s] for s in seq]
        except KeyError:
            raise GCGCAlphabetLetterEncodingException(f"{seq} not in {self.encoding_index}")

    def integer_decode(self, int_seq: Sequence[int]) -> str:
        """
        Given a sequence of integers, convert it to a string.

        Args:
            int_seq: A sequence of integers to convert.

        Returns:
            The string version of the integers list.
        """

        return "".join(self.decode_token(s) for s in int_seq)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved

from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Dict, List

from gcgc.fields.field import Field


@dataclass
class LabelField(Field):
    """
    A type of Field that can work with label values -- roundtripping them to an integer encoding.
    """

    encoding_dict: Dict[str, int]
    decoding_dict: Dict[int, str]

    def encode(self, label):
        """
        Look up the label in the encoding dict and return it.
        """

        return self.encoding_dict[label]

    def decode(self, label_int):
        """
        Look up the label's integer from the decoding dict and return it.
        """

        return self.decoding_dict[label_int]

    @classmethod
    def from_vocabulary(cls, name: str, vocab: List[str]):
        encoding_dict = {}
        decoding_dict = {}

        for i, s in enumerate(sorted(set(vocab))):
            encoding_dict[s] = i
            decoding_dict[i] = s

        return LabelField(name, encoding_dict, decoding_dict)


def default_preprocess(p: Path) -> str:
    return str(p)


@dataclass
class FileMetaDataField(LabelField):
    """
    """

    preprocess: Callable[[Path], str] = default_preprocess

    def encode(self, file: Path) -> int:
        """
        Preprocess the file path, then encode the resultant label.
        """
        label = self.preprocess(file)
        return super().encode(label)

    @classmethod
    def from_paths(
        cls, name: str, paths: List[Path], preprocess: Callable[[Path], str] = default_preprocess
    ):
        """
        Given a set of exemplar paths, create the (d-)encoding dict and return the field.
        """

        str_vocab = [preprocess(p) for p in paths]
        label_field = super().from_vocabulary(name, str_vocab)

        return cls(name, label_field.encoding_dict, label_field.decoding_dict, preprocess)


@dataclass
class AnnotationField(LabelField):
    """
    """

    preprocess: Callable[[Dict], str]

    def encode(self, file: Path) -> int:
        """
        Preprocess the file path, then encode the resultant label.
        """
        label = self.preprocess(file)
        return super().encode(label)

    @classmethod
    def from_annotations(
        cls, name: str, annotations: List[Dict], preprocess: Callable[[Dict], str]
    ):
        """
        Given a set of exemplar annotations, create the (d-)encoding dict and return the field.
        """

        str_vocab = [preprocess(a) for a in annotations]
        label_field = super().from_vocabulary(name, str_vocab)

        return cls(name, label_field.encoding_dict, label_field.decoding_dict, preprocess)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved

from gcgc.fields.categorical_field import LabelField, FileMetaDataField, AnnotationField

__all__ = ["LabelField", "FileMetaDataField", "AnnotationField"]



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['gcgc',
 'gcgc.alphabet',
 'gcgc.data',
 'gcgc.datasets',
 'gcgc.encoded_seq',
 'gcgc.fields',
 'gcgc.parser',
 'gcgc.tests',
 'gcgc.tests.alphabet',
 'gcgc.tests.cli',
 'gcgc.tests.datasets',
 'gcgc.tests.encoded_seq',
 'gcgc.tests.fields',
 'gcgc.tests.fixtures',
 'gcgc.tests.parser',
 'gcgc.tests.third_party.pytorch_utils',
 'gcgc.third_party',
 'gcgc.third_party.pytorch_utils']

package_data = \
{'': ['*'],
 'gcgc.data': ['splice/*'],
 'gcgc.tests.fixtures': ['ecoli/*', 'p53_human/*']}

install_requires = \
['aiohttp>=3.4,<4.0',
 'biopython==1.72',
 'numpy==1.15.2',
 'structlog>=18.2,<19.0']

extras_require = \
{'torch': ['torch==1.0.0']}

entry_points = \
{'console_scripts': ['gcgc = gcgc.cli:main']}

setup_kwargs = {
    'name': 'gcgc',
    'version': '0.3.0',
    'description': 'GCGC is a preprocessing library for biological sequence model development.',
    'long_description': '# GCGC\n\n> GCGC is a python package for feature processing on Biological Sequences.\n\n[![](https://img.shields.io/pypi/v/gcgc.svg)](https://pypi.python.org/pypi/gcgc)\n[![](https://img.shields.io/travis/tshauck/gcgc.svg)](https://travis-ci.org/tshauck/gcgc)\n\n## Installation\n\nInstall GCGC via pip:\n\n```sh\n$ pip install gcgc\n```\n\nIf you\'d like to use one of the third party tools, install the related "extras".\n\n```bash\n$ pip install gcgc -E torch\n```\n\n## Documentation\n\nThe GCGC documentation is at [gcgc.trenthauck.com](http://gcgc.trenthauck.com).\n\n## Development Board\n\nThe GCGC development board is hosted on [notion](https://www.notion.so/3649815c53324f01ae03abc99707dc68?v=98d8b29c39544dca9cde8ddc0dd8c98b).\n',
    'author': 'Trent Hauck',
    'author_email': 'trent@trenthauck.com',
    'url': 'http://gcgc.trenthauck.com/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved

from gcgc import alphabet
from gcgc.exceptions import GCGCAlphabetException


def biopython_alphabet_to_gcgc_alphabet(biopython_alphabet_instance):
    """
    Args:
        biopython_alphabet_instance
    """

    # Might also try just creating the type on the fly.

    for gcgc_alphabet in alphabet.__all__:

        klass = getattr(alphabet, gcgc_alphabet)
        instance = klass()

        if isinstance(instance, type(biopython_alphabet_instance)):
            return instance

    msg = f"No instance of {type(biopython_alphabet_instance)} among {alphabet.__all__}."
    raise GCGCAlphabetException(msg)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved

from pathlib import Path
from typing import Dict, Sequence

import torch
import torch.utils.data
from Bio import File, SeqIO

from gcgc.alphabet import ExtendedIUPACDNAEncoding
from gcgc.alphabet.base import EncodingAlphabet
from gcgc.parser import SequenceParser
from gcgc.parser.gcgc_record import GCGCRecord


class _SequenceIndexer(object):
    """
    This is a helper object that is used to mediate between the indexes in the underlying files
    and providing a consistent interface for __getitem__ in the GenomicDataset object.
    """

    def __init__(self):
        self._record_index = {}
        self._counter = -1

    def __call__(self, sid):
        try:
            return self._record_index[sid]
        except KeyError:
            self._counter = self._counter + 1
            self._record_index[sid] = self._counter
            return self._record_index[sid]


class GenomicDataset(torch.utils.data.Dataset):
    """
    GenomicDataset can be used to load sequence information into a format aminable to PyTorch.
    """

    def __init__(
        self,
        file_index: Dict[Path, File._IndexedSeqFileDict],
        parser: SequenceParser,
        file_format: str = "fasta",
    ):
        """
        Initialize the GenomicDataset object.

        It is recommended that you do _not_ itstantiate this object directly, and instead you should
        use one of the @classmethods.

        Args:
            file_index: A dictionary mapping the Path of the file to its index.
            parser: The parser object which specifies how the input sequence should be parsed.
            file_format: The format of the file to parse. This should be one understood by
                Biopython.
        """

        self._file_index = file_index
        self._parser = parser
        self._file_format = file_format

        super().__init__()

    @classmethod
    def from_path(
        cls,
        path: Path,
        parser: SequenceParser,
        file_format: str = "fasta",
        alphabet: EncodingAlphabet = ExtendedIUPACDNAEncoding(),
    ) -> "GenomicDataset":
        """
        Init from a single file. This is a convience method that delegates to
        from_paths.
        """
        return cls.init_from_path_generator([path], parser, file_format, alphabet)

    @classmethod
    def from_paths(
        cls,
        path_sequence: Sequence[Path],
        parser: SequenceParser,
        file_format="fasta",
        alphabet: EncodingAlphabet = ExtendedIUPACDNAEncoding(),
    ) -> "GenomicDataset":
        """
        Initialize the GenomicDataset from a pathlib.Path sequence. For example, the
        pathlib.Path.glob returns a Path generator.
        """

        file_index = {}
        si = _SequenceIndexer()

        for f in sorted(path_sequence):
            file_index[f] = SeqIO.index(str(f), file_format, key_function=si, alphabet=alphabet)

        return cls(file_index, parser=parser, file_format=file_format)

    def __len__(self) -> int:
        """
        Return the length of the dataset.
        """
        return sum(len(v) for v in self._file_index.values())

    def __getitem__(self, i: int):
        """
        Get the record from the index.
        """

        for k, v in self._file_index.items():
            try:
                r = GCGCRecord(path=k, seq_record=v[i])
                return self._parser.parse_record(r)
            except KeyError:
                pass

        else:
            raise RuntimeError(f"Exausted file index while looking for {i}.")



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved

import logging
import pathlib
import sys

import click

from gcgc import __version__
from gcgc.datasets import dataset

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


@click.group()
def main(args=None):
    """Console script for gcgc."""
    return 0


@main.command()
def version():
    print(__version__)


@main.command()
@click.argument("organism_id", nargs=-1)
@click.argument("data_directory", nargs=1)
def download_organism(organism_id, data_directory):
    path = pathlib.Path(data_directory)
    taxon_dataset = dataset.TaxonDataset(organism_id, path)
    taxon_dataset.download_files()


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover
